/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2011 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisISR.c - NDIS driver interrupt processing.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver.
   ---------------------------------------------------------------------------
*/


#include "target.h"
#include "NdisDevice.h"


/* -------------------------------------------------------------------------- */

/*
    ProcessReceive

Description:
    This function processes the buffer receiving.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (BOOLEAN):
    TRUE if a packet is received; FALSE otherwise.
*/

static
BOOLEAN ProcessReceive (
    IN  PNDIS_ADAPTER pAdapter,
    IN  int           cnLeft )
{
    int iNext;
    TDescStat status;
    PHARDWARE pHardware = &pAdapter->Hardware;
    PTDescInfo pInfo = &pHardware->m_RxDescInfo;
    PBUFFER_INFO pBufInfo = &pAdapter->RxBufInfo;
    PNET_BUFFER NetBuffer;
    PNET_BUFFER_LIST Nbl;
    PNET_BUFFER_LIST FirstPendingNbl = NULL;
    PNET_BUFFER_LIST LastPendingNbl = NULL;
    PNET_BUFFER_LIST FirstCopiedNbl = NULL;
    PNET_BUFFER_LIST LastCopiedNbl = NULL;
    PNET_BUFFER_LIST *FirstNbl;
    PNET_BUFFER_LIST *LastNbl;
    PTDesc pDesc;
    PDMA_BUFFER pDma;
    ULONG Flags;
    UINT i;
    UINT PendingPacketCount = 0;
    UINT CopiedPacketCount = 0;
    UINT *PacketCount;
    PUCHAR pBuffer;
    int nLength;
    int port = MAIN_PORT;

    iNext = pInfo->iNext;

    while (cnLeft--)
    {
        /* Get next descriptor which is not hardware owned. */
        GetReceivedPacket(pInfo, iNext, pDesc, status.ulData);
        if (status.rx.fHWOwned)
            break;

        pDesc->sw.Control.ulData = status.ulData;
        pInfo->pCurrent = pDesc;

        HardwareReceive(pHardware);

        if (pHardware->m_nPacketLen)
        {
            /* Get received packet information. */
            pDma = DMA_BUFFER(pDesc);

            Nbl = pDma->Nbl;
            NetBuffer = NET_BUFFER_LIST_FIRST_NB(Nbl);
            NET_BUFFER_DATA_LENGTH(NetBuffer) = pHardware->m_nPacketLen;

            /* Adjust the length of NDIS buffer to the length of the packet. */
            NdisAdjustBufferLength(pDma->Mdl, pHardware->m_nPacketLen);

            /* Flush the receive buffer. */
            NdisFlushBuffer(pDma->Mdl, FALSE);

            pBuffer = (PUCHAR) pDma->sm.pVirtual;

#ifdef DEBUG_RX_DATA
            for (nLength = 0; nLength < pHardware->m_nPacketLen; nLength++)
            {
                DbgPrint("%02X ", pBuffer[nLength]);
                if ((nLength % 16) == 15)
                {
                    DbgPrint("\n");
                }
            }
            DbgPrint("\n");
#endif
            nLength = pHardware->m_nPacketLen + 4;

            if (status.rx.fMulticast)
            {
                /* Descriptor does not have broadcast indication. */
                if (ETH_IS_BROADCAST(pBuffer))
                {
                    pHardware->m_cnCounter[port]
                        [OID_COUNTER_BROADCAST_FRAME_RCV]++;
                    pHardware->m_cnCounter[port]
                        [OID_COUNTER_BROADCAST_BYTES_RCV] += nLength;
                }
                else
                {
                    pHardware->m_cnCounter[port]
                        [OID_COUNTER_MULTICAST_FRAME_RCV]++;
                    pHardware->m_cnCounter[port]
                        [OID_COUNTER_MULTICAST_BYTES_RCV] += nLength;
                }
            }
            else
            {
                pHardware->m_cnCounter[port]
                    [OID_COUNTER_DIRECTED_FRAME_RCV]++;
                pHardware->m_cnCounter[port]
                    [OID_COUNTER_DIRECTED_BYTES_RCV] += nLength;
            }
            pHardware->m_cnCounter[port][OID_COUNTER_RCV_OK]++;

#if 0
            /* Get pointer of associated packet for further processing. */
            pPacket = pDma->pBuffer;

            if (pHardware->m_dwReceiveConfig &
                    (DMA_RX_CTRL_CSUM_IP | DMA_RX_CTRL_CSUM_TCP))
            {
                PNDIS_TCP_IP_CHECKSUM_PACKET_INFO pChecksum =
                    (PNDIS_TCP_IP_CHECKSUM_PACKET_INFO) &
                    NDIS_PER_PACKET_INFO_FROM_PACKET(pPacket,
                        TcpIpChecksumPacketInfo);

                if (pChecksum)
                {
                    pChecksum->Value = 0;

#if defined(CHECK_RCV_ERRORS) || defined(RCV_HUGE_FRAME)
                    if (status.rx.fCsumErrIP)
                        pChecksum->Receive.NdisPacketIpChecksumFailed =
                            TRUE;
                    else
#endif
                        pChecksum->Receive.NdisPacketIpChecksumSucceeded =
                            TRUE;

#if defined(CHECK_RCV_ERRORS) || defined(RCV_HUGE_FRAME)
                    if (status.rx.fCsumErrTCP)
                        pChecksum->Receive.NdisPacketTcpChecksumFailed =
                            TRUE;
                    else
#endif
                        pChecksum->Receive.NdisPacketTcpChecksumSucceeded =
                            TRUE;
#if 0
#if defined(CHECK_RCV_ERRORS) || defined(RCV_HUGE_FRAME)
                    if (status.rx.fCsumErrUDP)
                        pChecksum->Receive.NdisPacketUdpChecksumFailed =
                            TRUE;
                    else
#endif
                        pChecksum->Receive.NdisPacketUdpChecksumSucceeded =
                            TRUE;
#endif
                }
            }
#endif

            /* There are buffers available for using in descriptors. */
            if (NdisQueryDepthSList(&pBufInfo->listHead))
            {
                /* Upper layer can keep this packet. */
                NET_BUFFER_LIST_STATUS(Nbl) = NDIS_STATUS_SUCCESS;
                FirstNbl = &FirstPendingNbl;
                LastNbl = &LastPendingNbl;
                PacketCount = &PendingPacketCount;

                /* Reset this pointer so that we can debug the RX descriptor if
                   error happens.
                */
                DESC_FROM_RCV_NBL(Nbl) = NULL;

                /* Get a new buffer for this descriptor. */
                pDma = Alloc_DMA_Buffer(pBufInfo, pDesc);

                /* Set descriptor for packet receive. */
                SetReceiveBuffer(pDesc, pDma->sm.ulPhysical);
                SetReceiveLength(pDesc, pDma->ulSize);
                ReleasePacket(pDesc);
            }
            else
            {
                /* Upper layer must copy this packet immediately. */
                NET_BUFFER_LIST_STATUS(Nbl) = NDIS_STATUS_RESOURCES;
                FirstNbl = &FirstCopiedNbl;
                LastNbl = &LastCopiedNbl;
                PacketCount = &CopiedPacketCount;

                DESC_FROM_RCV_NBL(Nbl) = pDesc;

                /* Cannot release the descriptor yet. */
            }

            //
            // Add this NBL to the chain of NBLs to indicate up.
            //
            if (!*FirstNbl)
            {
                *LastNbl = *FirstNbl = Nbl;
            }
            else
            {
                NET_BUFFER_LIST_NEXT_NBL(*LastNbl) = Nbl;
                *LastNbl = Nbl;
            }
            (*PacketCount)++;
        }
        else
        {
            ReleasePacket(pDesc);
        }

        FreeReceivedPacket(pInfo, iNext);
    }
    pInfo->iNext = iNext;

    if (FirstPendingNbl)
    {
        /* We have packets to indicate. */
        Flags = NDIS_RECEIVE_FLAGS_DISPATCH_LEVEL;
#if 0
        Flags |= (NDIS_RECEIVE_FLAGS_PERFECT_FILTERED
#if defined(NDIS620_MINIPORT)
            | NDIS_RECEIVE_FLAGS_SINGLE_QUEUE
            //non-default queues use shared memory
            | (CurrentQueue?NDIS_RECEIVE_FLAGS_SHARED_MEMORY_INFO_VALID:0)
#endif
            );
#endif
        NET_BUFFER_LIST_NEXT_NBL(LastPendingNbl) = NULL;
        NdisMIndicateReceiveNetBufferLists(pAdapter->hAdapter,
            FirstPendingNbl, 0, PendingPacketCount, Flags);
    }
    if (FirstCopiedNbl)
    {
        /* We have packets to indicate. */
        Flags = NDIS_RECEIVE_FLAGS_DISPATCH_LEVEL;
#if 0
        Flags |= (NDIS_RECEIVE_FLAGS_PERFECT_FILTERED
#if defined(NDIS620_MINIPORT)
            | NDIS_RECEIVE_FLAGS_SINGLE_QUEUE
            //non-default queues use shared memory
            | (CurrentQueue?NDIS_RECEIVE_FLAGS_SHARED_MEMORY_INFO_VALID:0)
#endif
            );
#endif
        NDIS_SET_RECEIVE_FLAG(Flags, NDIS_RECEIVE_FLAGS_RESOURCES);
        NET_BUFFER_LIST_NEXT_NBL(LastCopiedNbl) = NULL;
        NdisMIndicateReceiveNetBufferLists(pAdapter->hAdapter,
            FirstCopiedNbl, 0, CopiedPacketCount, Flags);

        /* Determine which packets are kept for reuse and which ones are needed
           to put back to the list.
        */
        for (Nbl = FirstCopiedNbl; Nbl != NULL;
                Nbl = NET_BUFFER_LIST_NEXT_NBL(Nbl))
        {
            pDesc = DESC_FROM_RCV_NBL(Nbl);

            /* It is set by us so release the descriptor. */
            if (pDesc)
            {
                DESC_FROM_RCV_NBL(Nbl) = NULL;
                ReleasePacket(pDesc);
            }
        }
    }
    PendingPacketCount += CopiedPacketCount;
    return(PendingPacketCount != 0);
}  // ProcessReceive


/*
    MiniportHandleInterrupt

Description:
    This routine is used by NDIS for deferred interrupt processing.  It
    reads from the Interrupt Status Register any outstanding interrupts and
    handles them.

Parameters:
    NDIS_HANDLE hAdapaterContext
         Handle to adapter context containing adapter information.

Return (None):
*/

VOID MiniportHandleInterrupt (
    IN  NDIS_HANDLE hAdapterContext,
    IN  PVOID       pContext,
    IN  PVOID       ReceiveThrottle,
    IN  PVOID       Reserved2 )
{
    UINT IntEnable;
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PHARDWARE pHardware = &pAdapter->Hardware;
    int port = MAIN_PORT;
    int cnt = 0;
    int max_rcv = pHardware->m_RxDescInfo.cnAlloc;
    BOOLEAN rcv = FALSE;
#if defined(NDIS620_MINIPORT)
    PNDIS_RECEIVE_THROTTLE_PARAMETERS throttle = ReceiveThrottle;

    if (throttle->MaxNblsToIndicate != NDIS_INDICATE_ALL_NBLS)
    {
        max_rcv = throttle->MaxNblsToIndicate;
    }
#else
    UNREFERENCED_PARAMETER(ReceiveThrottle);
#endif
    UNREFERENCED_PARAMETER(pContext);
    UNREFERENCED_PARAMETER(Reserved2);

#ifdef DEBUG_DRIVER_INTERRUPT
DbgPrint("HandleInterrupt"NEWLINE);
#endif

    do
    {
        // Get the interrupt bits and save them.
        // Mask only interesting bits.
        HardwareReadInterrupt(pHardware, &IntEnable);
        if (!IntEnable)
        {
            break;
        }

#ifdef DEBUG_INTERRUPT
DbgPrint(" intr:%08x"NEWLINE, IntEnable);
#endif
        HardwareAcknowledgeInterrupt(pHardware, IntEnable);

        /* All packets have been sent. */
        if ((IntEnable & INT_TX_EMPTY))
        {
            /* Also acknowledge transmit complete interrupt. */
            HardwareAcknowledgeTransmit(pHardware);
            TransmitDone(pAdapter);

#ifdef DEBUG_COUNTER
            pHardware->m_nGood[COUNT_GOOD_INT_TX]++;
#endif
        }

        /* Do not need to process transmit complete as all transmitted
           descriptors are freed.
        */
        else
        if ((IntEnable & INT_TX))
        {
            TransmitDone(pAdapter);

#ifdef DEBUG_COUNTER
            pHardware->m_nGood[COUNT_GOOD_INT_TX]++;
#endif
        }

        if ((IntEnable & INT_RX))
        {

            // Receive the packet
            rcv = ProcessReceive(pAdapter, max_rcv);

#ifdef DEBUG_COUNTER
            pHardware->m_nGood[COUNT_GOOD_INT_RX]++;
#endif
        }

        if ((IntEnable & INT_PHY))
        {
            pHardware->m_bLinkIntWorking = TRUE;
            SwitchGetLinkStatus(pHardware);
            if (pAdapter->ulNdisMediaState !=
                    pHardware->m_ulHardwareState)
            {
                pAdapter->ulNdisMediaState = pHardware->m_ulHardwareState;
                IndicateLinkStatus(pAdapter);
            }
        }

        if ((IntEnable & INT_RX_OVERRUN))
        {

#ifdef DBG
            DBG_PRINT("Rx overrun"NEWLINE);
#endif
            pHardware->m_cnCounter[port][OID_COUNTER_RCV_NO_BUFFER]++;

            // Acknowledge the interrupt
            HardwareAcknowledgeOverrun(pHardware);
            HardwareResumeReceive(pHardware);

#ifdef DEBUG_COUNTER
            pHardware->m_nGood[COUNT_GOOD_INT_RX_OVERRUN]++;
#endif
        }

        if ((IntEnable & INT_RX_STOPPED))
        {

#ifdef DBG
            DBG_PRINT("Rx stopped"NEWLINE);
#endif

            /* Receive just has been stopped. */
            if (0 == pHardware->m_bReceiveStop) {
                pHardware->m_ulInterruptMask &= ~INT_RX_STOPPED;
            }
            else if (pHardware->m_bReceiveStop > 1) {
                if (pHardware->m_bEnabled  &&
                        (pHardware->m_dwReceiveConfig &
                        DMA_RX_CTRL_ENABLE)) {

#ifdef DBG
                    DBG_PRINT("Rx disabled"NEWLINE);
#endif
                    HardwareStartReceive(pHardware);
                }
                else {

#ifdef DBG
                    DBG_PRINT("Rx stop disabled:%d"NEWLINE,
                        pHardware->m_bReceiveStop);
#endif
                    pHardware->m_ulInterruptMask &= ~INT_RX_STOPPED;
                    pHardware->m_bReceiveStop = 0;
                }
            }

            /* Receive just has been started. */
            else {
                pHardware->m_bReceiveStop++;
            }
            HardwareAcknowledgeInterrupt(pHardware, INT_RX_STOPPED);
            break;
        }

#ifdef DBG
        if ((IntEnable & INT_TX_STOPPED))
        {
            UINT dwData;

            pHardware->m_ulInterruptMask &= ~INT_TX_STOPPED;
            DBG_PRINT("Tx stopped"NEWLINE);
            HW_READ_DWORD(pHardware, REG_DMA_TX_CTRL, &dwData);
            if (!(dwData & DMA_TX_CTRL_ENABLE)) {
                DBG_PRINT("Tx disabled"NEWLINE);
            }
            HardwareAcknowledgeInterrupt(pHardware, INT_TX_STOPPED);
            break;
        }
#endif
    } while (1);

#if defined(NDIS620_MINIPORT)
    if (rcv)
    {
        /* This routine will be called again until no more receive. */
        throttle->MoreNblsPending = 1;
        return;
    }
#endif

#ifdef DEBUG_INTR_LOCK
    while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
        InterlockedDecrement(&pAdapter->OperateAcquire);
    pAdapter->InterruptOperate |= INTR_ACQ_HANDLE << 16;
    InterlockedDecrement(&pAdapter->OperateAcquire);
#endif
    if (1 != InterlockedIncrement(&pAdapter->InterruptAcquire))
    {
#ifdef DEBUG_INTR_LOCK
        if (pAdapter->InterruptOperate & INTR_ACQ_ISR)
        {
            DbgPrint(" checking intr: %x\n", pAdapter->InterruptOperate);
        }
        else
#endif
        {
            InterlockedDecrement(&pAdapter->InterruptAcquire);
            AcquireInterrupt(pAdapter, INTR_ACQ_HANDLE);
        }
    }
#ifdef DEBUG_INTR_LOCK
    while (1 != InterlockedIncrement(&pAdapter->OperateAcquire))
        InterlockedDecrement(&pAdapter->OperateAcquire);
    pAdapter->InterruptOperate |= INTR_ACQ_HANDLE;
    pAdapter->InterruptOperate &= ~INTR_ACQ_ISR;
    InterlockedDecrement(&pAdapter->OperateAcquire);
#endif
    HardwareRestoreInterrupt_PCI(pHardware, pHardware->m_ulInterruptMask);
    ReleaseInterrupt(pAdapter, INTR_ACQ_HANDLE);
}  // MiniportHandleInterrupt


/*
    MiniportISR

Description:
    This routine is used by NDIS for the hardware to check if the interrupt
    is its own in the shared interrupt environment.

Parameters:
    NDIS_HANDLE hAdapaterContext
         Handle to adapter context containing adapter information.

    PBOOLEAN pbQueueDpc
        Buffer to indicate the deferred interrupt processing should be
        scheduled.

    PULONG TargetProcessors
        Bitmap to indicate which processor the deferred processing should run.

Return (BOOLEAN):
    PBOOLEAN pbInterruptRecognized
        Buffer to indicate the interrupt is recognized

*/

BOOLEAN MiniportISR (
    IN  NDIS_HANDLE hAdapterContext,
    OUT PBOOLEAN    pbQueueDpc,
    OUT PULONG      TargetProcessors )
{
    BOOLEAN bInterruptRecognized;
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PHARDWARE pHardware = &pAdapter->Hardware;
    UINT ulIntStatus = 0;

    if (1 != InterlockedIncrement(&pAdapter->InterruptAcquire))
    {
#ifdef DEBUG_INTR_LOCK
        if (pAdapter->InterruptOperate &
               (INTR_ACQ_CHECK | (INTR_ACQ_CHECK << 16)))
            pAdapter->InterruptOperate |= INTR_ACQ_ISR;
#if 1
        else if (!(pAdapter->InterruptOperate & INTR_ACQ_HANDLE))
            DBG_PRINT("isr: %x\n", pAdapter->InterruptOperate);
#endif
#endif
    }
    HardwareReadInterrupt(pHardware, &ulIntStatus);

    /* Hardware interrupts are disabled. */
    ulIntStatus &= pHardware->m_ulInterruptSet;

    /* Got interrupt and not already queued for interrupt handling. */
    if (ulIntStatus)
    {
        HardwareBlockInterrupt_PCI(pHardware);

        bInterruptRecognized = TRUE;
        *pbQueueDpc = TRUE;
    }
    else
    {
        // not our interrupt!
        bInterruptRecognized = FALSE;
        *pbQueueDpc = FALSE;
        *TargetProcessors = 0;
    }
    InterlockedDecrement(&pAdapter->InterruptAcquire);
    return bInterruptRecognized;
}  // MiniportISR

/* -------------------------------------------------------------------------- */

/*
    MiniportReturnNetBufferLists

Description:
    This routine returns the net buffer lists passed to NDIS.

    NDIS Miniport entry point called whenever protocols are done with one or
    NBLs that we indicated up with NdisMIndicateReceiveNetBufferLists.

    Note that the list of NBLs may be chained together from multiple separate
    lists that were indicated up individually.

Parameters:
    NDIS_HANDLE hAdapaterContext
        Handle to adapter context containing adapter information.

    PNET_BUFFER_LIST NetBufferLists
        Pointer to NDIS net buffer lists.

    ULONG ReturnFlags
        Indicate dispatch level.

        May contain the NDIS_RETURN_FLAGS_DISPATCH_LEVEL
        flag, which if is set, indicates we can get a
        small perf win by not checking or raising the
        IRQL

Return (None):
*/

VOID MiniportReturnNetBufferLists (
    IN  NDIS_HANDLE      hAdapterContext,
    IN  PNET_BUFFER_LIST NetBufferLists,
    IN  ULONG            ReturnFlags )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PDMA_BUFFER pDma;

#ifdef DEBUG_DRIVER_RCV
    DBG_PRINT("ReturnNetBufferLists"NEWLINE);
#endif

    while (NetBufferLists)
    {
        pDma = DMA_FROM_RCV_NBL(NetBufferLists);
        Free_DMA_Buffer(&pAdapter->RxBufInfo, pDma);
        NetBufferLists = NET_BUFFER_LIST_NEXT_NBL(NetBufferLists);
    }
}  // MiniportReturnNetBufferLists
